import os
import threading

class LazyLoadedAPI:
    """
    LazyLoadedAPI 클래스로 다른 클래스 선언부를 감싸서, 객체의 attr 에 접근할때 초기화할 수 있도록 합니다.
    """
    def __init__(self, api_class, *args, **kwargs):
        self.api_class = api_class
        self.args = args
        self.kwargs = kwargs
        self._instance = None
        self._lock = threading.Lock()

    def _initialize_instance(self):
        if self._instance is None:
            with self._lock:
                if self._instance is None:
                    self._instance = self.api_class(*self.args, **self.kwargs)

    def __getattr__(self, name):
        self._initialize_instance()
        return getattr(self._instance, name)



from sawsi.aws import shared
from typing import Any


class LambdaAPI:
    def __init__(self, credentials=None, region=shared.DEFAULT_REGION):
        self.boto3_session = shared.get_boto_session(credentials)
        self.lambda_client = self.boto3_session.client('lambda', region_name=region)

    def invoke(self, function_name: str, payload: Any):
        response = self.lambda_client.invoke(
            FunctionName=function_name,
            InvocationType='RequestResponse',  # 'Event' for asynchronous execution
            Payload=payload
        )

        # 응답에서 Payload 추출
        response_payload = response['Payload'].read()
        response_body = response_payload.decode('utf-8')
        return response_body

    def create_event_source_mapping(
            self, function_name: str, event_source_arn: str,
            enabled: bool = True, batch_size: int = 100, starting_position: str = 'LATEST'
    ):
        """
        Lambda 함수에 트리거 추가
        """
        response = self.lambda_client.create_event_source_mapping(
            EventSourceArn=event_source_arn,
            FunctionName=function_name,
            Enabled=enabled,
            BatchSize=batch_size,  # 처리할 항목 수 조정 가능
            StartingPosition=starting_position  # 최신 항목부터 시작
        )
        return response



import os
import base64
import hashlib


def make_salt():
    # 무작위 salt 값 생성
    salt = base64.b64encode(os.urandom(16))
    return salt


def hash_password(plain):
    return hashlib.sha512(plain.encode('utf-8')).hexdigest()


def create_hash(password, salt=None):
    # Generate the salt.
    PBKDF2_COMPAT_SALT_BYTES = 24
    if not salt:
        salt = ''

        if os.path.exists('/dev/urandom'):
            with open('/dev/urandom', 'rb') as f:
                salt = base64.b64encode(f.read(PBKDF2_COMPAT_SALT_BYTES)).decode('utf-8')
        else:
            for i in range(0, PBKDF2_COMPAT_SALT_BYTES, 2):
                salt += chr(os.urandom(2)[1])

            salt = base64.b64encode(salt.encode()).decode('utf-8')

    # Determine the best supported algorithm and iteration count.
    PBKDF2_COMPAT_HASH_ALGORITHM = 'sha256'
    PBKDF2_COMPAT_ITERATIONS = 12000
    PBKDF2_COMPAT_HASH_BYTES = 24

    algo = PBKDF2_COMPAT_HASH_ALGORITHM.lower()
    iterations = PBKDF2_COMPAT_ITERATIONS

    # Return format: algorithm:iterations:salt:hash
    algorithm = 'sha256'
    pbkdf2 = hash_pbkdf2(algorithm, password, salt, iterations, PBKDF2_COMPAT_HASH_BYTES, True)
    # pbkdf2 = hashlib.pbkdf2_hmac(algo, password.encode(), base64.b64decode(salt), iterations, PBKDF2_COMPAT_HASH_BYTES)
    prefix = algo if algo else 'sha1'

    return f"{prefix}:{iterations}:{salt}:{base64.b64encode(pbkdf2).decode('utf-8')}"


def hash_pbkdf2(algo, password, salt, iterations, length=0, binary=False):
    if not binary:
        length = length * 2  # Convert length from bytes to hex representation length
    dk = hashlib.pbkdf2_hmac(algo, password.encode(), salt.encode(), iterations, length)
    if binary:
        return dk
    return dk.hex()


def validate_password(password, hash):
    # Split the hash into 4 parts.
    tokens = hash.split(':')
    algo = tokens[0]
    rounds = tokens[1]
    salt = tokens[2]
    hv = tokens[3]
    return create_hash(password, salt) == hash


def slow_equals(a, b):
    # Compare two byte strings in a time-constant manner.
    if len(a) != len(b):
        return False

    result = 0
    for x, y in zip(a, b):
        result |= x ^ y
    return result == 0


if __name__ == '__main__':
    # sat = 'SALT'
    p = create_hash('Kch704467!', salt='vOfePPE220ccUcEHe2TOUdX9S7XXVx40')
    print(p)
    salt = 'vOfePPE220ccUcEHe2TOUdX9S7XXVx40'
    iterations = 12000
    hash_length = 24


    # print(len(p))
    # v = validate_password('123456', p)
    # print(v)

    v = validate_password('Kch704467!', 'sha256:12000:vOfePPE220ccUcEHe2TOUdX9S7XXVx40:t6iYQs9YscWr2ck+EFK4MwCNecVAApXc')
    print(v)


"""
list 객체 관련 유틸
"""


def divide_chunks(l, n):
    # 입력된 리스트 l 을 n 개씩 갖도록 split
    # looping till length l
    if isinstance(l, list):
        for i in range(0, len(l), n):
            yield l[i:i + n]
    elif isinstance(l, dict):
        items = list(l.items())
        for i in range(0, len(items), n):
            chunks = items[i: i+n]
            dic = {}
            for key, value in chunks:
                dic[key] = value
            yield dic



import tempfile
from sawsi.aws import shared


class Firehose:
    def __init__(self, boto3_session, region=shared.DEFAULT_REGION):
        self.client = boto3_session.client('firehose', region_name=region)
        self.region = region


    def create_delivery_stream(self, delivery_stream_name, role_arn, s3_bucket_name, object_key_prefix, error_prefix='error_logs'):
        """
        Kinesis Firehose Delivery Stream 생성
        :param delivery_stream_name: 스트림 이름
        :param role_arn: S3에 쓸때 필요한 권한
        :param s3_bucket_name: 저장할 S3 이름
        :param object_key_prefix: 저장시 객체 이름 프리픽스 (EX: "item")
        :param error_prefix: 에러 발생시 객체 이름 프리픽스
        :return:
        """
        firehose_response = self.client.create_delivery_stream(
            DeliveryStreamName=delivery_stream_name,
            S3DestinationConfiguration={
                'RoleARN': role_arn,
                'BucketARN': 'arn:aws:s3:::{}'.format(s3_bucket_name),
                'Prefix': f'{object_key_prefix}/',  # 필요에 따라 수정
                'ErrorOutputPrefix': f'{error_prefix}/'  # 필요에 따라 수정
            }
        )
        return firehose_response


    def put_record(self, delivery_stream_name, data):
        """
        :param delivery_stream_name:
        :param data: Data can be binary, dict, str, ...any
        :return:
        """
        response = self.client.put_record(
            DeliveryStreamName=delivery_stream_name,
            Record={
                'Data': data
            }
        )
        return response


import json

import urllib3
from sawsi.aws import shared
from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from sawsi.shared.dict_util import convert_dynamodb_document_to_dict
from typing import Literal

http = urllib3.PoolManager()

service = 'es'

class OpenSearchAPI:
    """
    API Gateway Management API
    """
    def __init__(self, domain_endpoint_url, credentials=None, region=shared.DEFAULT_REGION):
        """
        :param credentials: {
            "aws_access_key_id": "str",
            "aws_secret_access_key": "str",
            "region_name": "str",
            "profile_name": "str",
        }
        """
        self.boto3_session = shared.get_boto_session(credentials)
        self.cache = {}
        self.domain_endpoint_url = domain_endpoint_url
        if self.domain_endpoint_url.endswith('/'):
            # 뒤에 슬레쉬 짜름
            self.domain_endpoint_url = self.domain_endpoint_url[:-1]
        self.region = region

    def _http_request(self, method, url, body, region):
        signed_request = self._sign_request(method, url, body=body, region=region)
        response = http.request(
            method,
            signed_request.url,
            body=signed_request.body,
            headers=signed_request.headers
        )
        # status = response.status
        # response_body = response.data.decode('utf-8')
        return response

    def _sign_request(self, method, url, body, region):
        # BOTO3 서명
        credentials = self.boto3_session.get_credentials()
        headers = {'Content-Type': 'application/json'}
        request = AWSRequest(method=method, url=url, data=body, headers=headers)
        SigV4Auth(credentials, service, region).add_auth(request)
        return request.prepare()

    def run(self, method: Literal["GET", "POST", "PUT", "DELETE"], path, body):
        url = f'{self.domain_endpoint_url}/{path}'
        response = self._http_request(method, url, body, region=self.region)
        return response

    def put_document(self, index_name: str, doc_id: str, document: dict):
        path = f"{index_name}/_doc/{doc_id}"
        doc_body = convert_dynamodb_document_to_dict(document)
        doc_body = json.dumps(doc_body)
        response = self.run('PUT', path, doc_body)
        return response

    def delete_document(self, index_name: str, doc_id: str):
        path = f"{index_name}/_doc/{doc_id}"
        response = self.run('DELETE', path, None)
        return response

    def bulk_put_documents(self, index_name: str, documents: dict[str, dict]):
        """
        여러 문서를 한번에 인덱싱하는 메서드
        :param index_name: 인덱스 이름
        :param documents: {"doc_id1": document1, "doc_id2": document2, ...}
        """
        bulk_data = ""
        for doc_id, document in documents.items():
            doc_body = convert_dynamodb_document_to_dict(document)
            bulk_data += json.dumps({"index": {"_index": index_name, "_id": doc_id}}) + "\n"
            bulk_data += json.dumps(doc_body) + "\n"

        response = self.run('POST', '_bulk', bulk_data)
        return response

    def bulk_delete_documents(self, index_name: str, doc_ids: list):
        """
        여러 문서를 한번에 삭제하는 메서드
        :param index_name: 인덱스 이름
        :param doc_ids: 삭제할 문서의 ID 리스트
        """
        bulk_data = ""
        for doc_id in doc_ids:
            bulk_data += json.dumps({"delete": {"_index": index_name, "_id": doc_id}}) + "\n"

        response = self.run('POST', '_bulk', bulk_data)
        return response



from sawsi.aws.shared import DEFAULT_REGION


class SecretManager:
    def __init__(self, boto3_session, region=DEFAULT_REGION):
        self.client = boto3_session.client('secretsmanager', region_name=region)
        self.region = boto3_session.region_name

    def get_secret_value_response(self, secret_name):
        response = self.client.get_secret_value(
            SecretId=secret_name
        )
        # Decrypts secret using the associated KMS key.
        secret = response['SecretString']
        return secret
